import numpy as np 

def _detect_shape(arr):
    start_point = arr[0,0]
    diff = arr[1,0] - arr[0,0]
    idx1 = 0
    print(start_point,diff)
    for i,row in enumerate(arr):
        print(row[0])
        if np.isclose(row[0],start_point,rtol=diff*0.01) and i > 0:
            idx1 = i
            break
    
    return (idx1,int(arr.shape[0]/(idx1)),3)


def _readin(filename,sep=',',header=True):
    headers = [' ']
    idx_offset = 0
    with open(filename,'r') as file:
        lines = file.read().splitlines()

    if header:
        headers = [col for col in lines[0].split(sep)]
        # remove '\n'
        headers[-1] = headers[-1][:-1]
        lines = lines[1:]

    linelengths = [len(l.split(sep)) for l in lines]

    if all(np.isclose(linelengths,3)):
        print('csv is of suitable format')
    elif all(np.isclose(linelengths,4)) and (headers[0] == 'idx' or headers[0] == 'index'):
        idx_offset = 1
    else:
        print("Not all rows have 3 entries.")
        quit()

    arr = np.empty((len(lines),3)) 

    for i,l in enumerate(lines):
        elements = l.split(sep)
        arr[i,0] = float(elements[0+idx_offset])
        arr[i,1] = float(elements[1+idx_offset])
        arr[i,2] = float(elements[2+idx_offset])
        
    return arr
    
        

def convert(filename,outputname,sep=',',header=True):
    '''
    Converts a csv called `filename` to a text file `outputname` which can be used to produce a pgfplot in latex of the parameter space.
    '''
    arr = _readin(filename,sep,header)

    shape = _detect_shape(arr)

    arr = arr.reshape(shape)

    with open(outputname,'w') as output:
        for block in arr:
            for row in block:
                for item in row:
                    output.write(f'{item} ') 
                output.write('\n')
            output.write('\n')


from setuptools import setup, find_packages


with open('README.md', encoding='utf-8') as f:
    readme = f.read()

setup(
    name='bulksimulator',
    version='0.3.2',
    description='Module to run a simulation for many parameter choices',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='Jeremy Worsfold',
    author_email='jw3286@bath.ac.uk',
    url='https://github.bath.ac.uk/jw3286/bulksimulator',
    license="GPL3",
    packages=find_packages(exclude=('tests'))
)


import multiprocessing as mp
import numpy as np
import itertools
from functools import partial

class BulkSimulator:
    '''
    This class iterates over a parameter space, simulating for each set of parameters

    BulkSimulator(simulation,params,procs,kwargs,returnnames=['Result'])

    simulation: function which carries out the simulation

    params: dictionary of parameters with the names of the parameters and list containing values to be iterated over

    procs: number of processes to be ran on

    kwargs: other parameters

    '''
    def __init__(self,simulation,params,procs,kwargs,returnnames=['Result']):
        self.simulation = simulation
        self.kwargs = kwargs
        self.params = list(params.values())
        self.procs = procs
        self.param_names = params.keys()
        self.returnnames = returnnames

        grid = np.meshgrid(*self.params)
        self.gridsize = list(np.array(grid).shape)
        grid = tuple([gr.ravel() for gr in list(grid)])
        mesh = np.c_[grid]
        self.params = []
        
        self.params = [list(col) for col in mesh.T]
        self.N = len(self.params[0])
        self.args = [[val] * self.N for val in self.kwargs.values()]
        self.args = self.params + self.args 
        self.args = list(zip(*tuple(self.args)))  


    def run(self):
        '''
        This method runs the simulation for all parameters choices chosen on the number of processes speficied
        '''
        with mp.Pool(processes=self.procs) as pool:
            self.results = pool.starmap(self.simulation, self.args)
        if not isinstance(self.results, tuple):
            self.results = (self.results,)
    

    def to_csv(self,filename,sep=','):
        '''
        Given a filename and optional separator character, this function saves the parameters and results to a csv file.
        '''
        with open(filename,'w') as file:
            header = [f'idx{sep}']
            for name in self.param_names:
                header.append(f'{name}{sep}')
            for name in self.returnnames:
                header.append(f'{name}{sep}')
            header_str = ''.join(header)[:-1]
            file.write(header_str+'\n')

            for i,row in enumerate(zip(*(self.params+list(self.results)))):
                param_txt = [f'{val}{sep}' for val in row]
                row_txt = f'{i}{sep}' + ''.join(param_txt)[:-1] + '\n'
                file.write(row_txt)


    def save_metadata(self,filename,sep=','):
        '''
        Saves the extra "kwargs" to a csv file so all settings for a run can be saved
        '''
        with open(filename,'w') as file:
            header = []
            for name in self.kwargs.keys():
                header.append(f'{name}{sep}')
            header_str = ''.join(header)[:-1]
            file.write(header_str+'\n')

            line = [f'{val}{sep}' for val in self.kwargs.values()]
            line_str = ''.join(line)[:-1]
            file.write(line_str+'\n')


    def __str__(self):
        header = ['idx,']
        for name in self.param_names:
            header.append(f'{name},')
        for name in self.returnnames:
            header.append(f'{name},')
        header_str = ''.join(header)[:-1]
        row_txt = []
        for i,row in enumerate(zip(*(self.params+list(self.results)))):
            param_txt = [f'{val},' for val in row]
            row_txt.append(f'{i},' + ''.join(param_txt)[:-1] + '\n')
        res_string = ''.join(row_txt)
        return f'''
simulation = {self.simulation.__name__}
procs = {self.procs}
kwargs: {self.kwargs}
number of results: {self.N}
{header_str}
{res_string}'''